import matplotlib.pyplot as plt  # needed for visual inspection
import numpy as np
import pandas as pd
import pytest

from locan import (
    Ellipse,
    EmptyRegion,
    Interval,
    LocData,
    Polygon,
    Rectangle,
    add_drift,
    resample,
    simulate_frame_numbers,
    simulate_tracks,
)
from locan.simulation import (
    make_cluster,
    make_dstorm,
    make_Matern,
    make_NeymanScott,
    make_Poisson,
    make_Thomas,
    make_uniform,
    simulate_cluster,
    simulate_dstorm,
    simulate_Matern,
    simulate_NeymanScott,
    simulate_Poisson,
    simulate_Thomas,
    simulate_uniform,
)
from locan.simulation.simulate_drift import _drift, _random_walk_drift


def test_make_uniform():
    rng = np.random.default_rng(seed=1)
    samples = make_uniform(n_samples=10, region=EmptyRegion(), seed=rng)
    assert np.size(samples) == 0
    samples = make_uniform(n_samples=10, region=(0, 1), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples = make_uniform(n_samples=10, region=((0, 1), (10, 11)), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples = make_uniform(
        n_samples=10, region=((0, 1), (10, 11), (100, 101)), seed=rng
    )
    assert len(samples) == 10
    assert samples.shape[1] == 3
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples = make_uniform(n_samples=10, region=Interval(0, 1), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples = make_uniform(n_samples=10, region=Rectangle((0, 10), 1, 1, 0), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples = make_uniform(n_samples=10, region=Rectangle((0, 10), 1, 1, 45), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(-0.71 <= samples[:, 0])  # 0.71 ~ (np.sqrt(2) / 2)
    assert np.all(samples[:, 0] < 0.71)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < (10 + 1.42))

    samples = make_uniform(n_samples=10, region=Ellipse((0, 0), 1, 1, 0), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(-0.5 <= samples[:, 0])
    assert np.all(samples[:, 0] < 0.5)
    assert np.all(-0.5 <= samples[:, 1])
    assert np.all(samples[:, 1] < 0.5)

    samples = make_uniform(n_samples=10, region=Ellipse((0, 0), 1, 2, 0), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(-0.5 <= samples[:, 0])
    assert np.all(samples[:, 0] < 0.5)
    assert np.all(-1 <= samples[:, 1])
    assert np.all(samples[:, 1] < 1)

    samples = make_uniform(
        n_samples=10, region=Polygon(((0, 10), (1, 11), (1, 10))), seed=rng
    )
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)


def test_simulate_uniform():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_uniform(n_samples=10, region=(0, 1), seed=rng)
    assert len(locdata) == 10
    assert locdata.dimension == 1
    assert repr(locdata.region) == "Interval(0, 1)"


def test_make_Poisson():
    rng = np.random.default_rng(seed=1)

    samples = make_Poisson(intensity=10, region=EmptyRegion(), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples = make_Poisson(intensity=1e-10, region=(0, 1), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples = make_Poisson(intensity=1e-10, region=((0, 1), (10, 11)), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 3

    samples = make_Poisson(intensity=10, region=(0, 1), seed=rng)
    assert samples.shape[1] == 1
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples = make_Poisson(intensity=10, region=((0, 1), (10, 11)), seed=rng)
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples = make_Poisson(
        intensity=10, region=((0, 1), (10, 11), (100, 101)), seed=rng
    )
    assert samples.shape[1] == 3
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples = make_Poisson(intensity=10, region=Interval(0, 1), seed=rng)
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    samples = make_Poisson(intensity=10, region=Rectangle((0, 10), 1, 1, 0), seed=rng)
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples = make_Poisson(intensity=10, region=Ellipse((0, 0), 1, 1, 0), seed=rng)
    assert samples.shape[1] == 2
    assert np.all(-0.5 <= samples[:, 0])
    assert np.all(samples[:, 0] < 0.5)
    assert np.all(-0.5 <= samples[:, 1])
    assert np.all(samples[:, 1] < 0.5)

    samples = make_Poisson(intensity=10, region=Ellipse((0, 0), 1, 2, 0), seed=rng)
    assert samples.shape[1] == 2
    assert np.all(-0.5 <= samples[:, 0])
    assert np.all(samples[:, 0] < 0.5)
    assert np.all(-1 <= samples[:, 1])
    assert np.all(samples[:, 1] < 1)

    samples = make_Poisson(
        intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))), seed=rng
    )
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)


@pytest.mark.visual
def test_make_Poisson_visual():
    rng = np.random.default_rng(seed=1)

    samples = make_Poisson(
        intensity=1000, region=Polygon(((0, 10), (1, 11), (1, 10))), seed=rng
    )

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis("equal")
    plt.show()


def test_simulate_Poisson():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_Poisson(intensity=10, region=(0, 1), seed=rng)
    assert locdata.dimension == 1
    assert repr(locdata.region) == "Interval(0, 1)"


def test_make_cluster():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_cluster(
        centers=10, region=EmptyRegion(), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_cluster(
        centers=0, region=(0, 1), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=(0, 1),
        expansion_distance=1,
        offspring=None,
        clip=False,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert len(parent_samples) == 10
    assert np.all(-1 <= samples[:, 0])
    assert np.all(samples[:, 0] < 2)

    samples, labels, parent_samples, region = make_cluster(
        centers=(-1, 0, 0.5, 1, 2),
        region=(0, 1),
        expansion_distance=1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert len(parent_samples) == 5
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    offspring = [np.linspace(-0.1, 0.1, 11)] * 20
    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=(0, 1),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert len(parent_samples) == 10
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    with pytest.raises(TypeError):
        offspring = [np.linspace(-0.1, 0.1, 11)] * 2
        make_cluster(
            centers=10,
            region=(0, 1),
            expansion_distance=1,
            offspring=offspring,
            clip=True,
            shuffle=True,
            seed=rng,
        )

    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=((0, 1), (10, 11), (100, 101)),
        expansion_distance=0.1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert len(parent_samples) == 10
    assert samples.ndim == 2
    assert samples.shape[1] == 3
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=Interval(0, 1),
        expansion_distance=1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert len(parent_samples) == 10
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=Rectangle((0, 10), 1, 1, 0),
        expansion_distance=1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert len(parent_samples) == 10
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_cluster(
        centers=((-0.5, 10), (0.5, 11), (1, 0)),
        region=Rectangle((0, 10), 1, 1, 0),
        expansion_distance=1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert len(parent_samples) == 3
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] <= 11)

    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_distance=1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert len(parent_samples) == 10
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    def offspring(parent):
        angles = np.linspace(0, 360, 36)
        for angle in angles:
            circle = Ellipse(parent, 1, 1, angle)
        return circle.points

    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    def offspring(parent):
        from math import sin

        return [sin(parent)]

    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=(0, 1),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    offspring = [
        make_uniform(n_samples=20, region=Ellipse((0, 0), 0.1, 0.1), seed=rng)
    ] * 10
    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_distance=0.1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    offspring = [
        np.array([[1, 2], [1, 3], [2, 3]]),
        np.array([]),
        np.array([[1, 2], [1, 3], [2, 3]]),
    ]
    samples, labels, parent_samples, region = make_cluster(
        centers=3,
        region=((0, 10), (0, 10)),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)
    assert np.all(0 <= samples[:, 1])
    assert np.all(samples[:, 1] < 10)

    offspring = [np.array([])] * 20
    samples, labels, parent_samples, region = make_cluster(
        centers=3,
        region=((0, 10), (0, 10)),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == 0
    assert samples.ndim == 2


@pytest.mark.visual
def test_make_cluster_visual():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_cluster(
        centers=10,
        region=(0, 1),
        expansion_distance=1,
        offspring=[np.linspace(-0.1, 0.1, 11)] * 10,
        clip=False,
        shuffle=True,
        seed=rng,
    )

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples, [0] * len(samples), alpha=0.1)
    plt.scatter(parent_samples, [0] * len(parent_samples), color="Red", alpha=0.1)
    ax.axis("equal")
    plt.show()

    def offspring_points(parent):
        angles = np.linspace(0, 360, 36)
        for angle in angles:
            circle = Ellipse(parent, 1, 1, angle)
        return circle.points

    samples, labels, parent_samples, region = make_cluster(
        centers=5,
        region=Polygon(((0, 10), (10, 20), (10, 10))),
        offspring=offspring_points,
        clip=True,
        shuffle=True,
        seed=rng,
    )

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis("equal")
    plt.show()


def test_simulate_cluster():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_cluster(
        centers=10,
        region=(0, 1),
        expansion_distance=1,
        offspring=None,
        clip=False,
        shuffle=False,
        seed=rng,
    )
    assert locdata.dimension == 1
    assert repr(locdata.region) == "Interval(-1, 2)"
    assert "cluster_label" in locdata.data.columns


def test_make_NeymanScott():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10, region=EmptyRegion(), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=1e-10, region=(0, 1), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=(0, 1),
        expansion_distance=1,
        offspring=None,
        clip=False,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-1 <= samples[:, 0])
    assert np.all(samples[:, 0] < 2)

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=(0, 1),
        expansion_distance=0,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    offspring = [np.linspace(-0.1, 0.1, 11)] * 200
    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=(0, 1),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-1 <= samples[:, 0])
    assert np.all(samples[:, 0] < 2)

    with pytest.raises(TypeError):
        offspring = [np.linspace(-0.1, 0.1, 11)] * 2
        make_NeymanScott(
            parent_intensity=10,
            region=(0, 1),
            expansion_distance=1,
            offspring=offspring,
            clip=True,
            shuffle=True,
            seed=rng,
        )

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=((0, 1), (10, 11), (100, 101)),
        expansion_distance=1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 3
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=Interval(0, 1),
        expansion_distance=1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=Rectangle((0, 10), 1, 1, 0),
        expansion_distance=1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_distance=1,
        offspring=None,
        clip=False,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.shape[1] == 2
    assert samples.ndim == 2
    assert np.all(-1 <= samples[:, 0])
    assert np.all(samples[:, 0] < 2)
    assert np.all(9 <= samples[:, 1])
    assert np.all(samples[:, 1] < 12)

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_distance=1,
        offspring=None,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.shape[1] == 2
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    def offspring(parent):
        angles = np.linspace(0, 360, 36)
        for angle in angles:
            circle = Ellipse(parent, 1, 1, angle)
        return circle.points

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.shape[1] == 2
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    def offspring(parent):
        from math import sin

        return [sin(parent)]

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=10,
        region=(0, 1),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.shape[1] == 1
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    offspring = [
        np.array([[1, 2], [1, 3], [2, 3]]),
        np.array([]),
        np.array([[1, 2], [1, 3], [2, 3]]),
    ] * 20
    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=0.1,
        region=((0, 10), (0, 10)),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.shape[1] == 2
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    offspring = [np.array([])] * 20
    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=0.1,
        region=((0, 10), (0, 10)),
        expansion_distance=1,
        offspring=offspring,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == 0
    assert samples.ndim == 2


@pytest.mark.visual
def test_make_NeymanScott_visual():
    rng = np.random.default_rng(seed=1)

    p_intensity = 0.5
    region = Polygon(((0, 10), (10, 20), (10, 10)))

    def offspring_points(parent):
        angles = np.linspace(0, 360, 36)
        for angle in angles:
            circle = Ellipse(parent, 1, 1, angle)
        return circle.points

    samples, labels, parent_samples, region = make_NeymanScott(
        parent_intensity=p_intensity,
        region=region,
        expansion_distance=1,
        offspring=offspring_points,
        clip=True,
        shuffle=True,
        seed=rng,
    )

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis("equal")
    plt.show()


def test_simulate_NeymanScott():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_NeymanScott(
        parent_intensity=10,
        region=(0, 1),
        expansion_distance=0,
        offspring=None,
        clip=False,
        shuffle=True,
        seed=rng,
    )
    assert locdata.dimension == 1
    assert repr(locdata.region) == "Interval(0, 1)"
    assert "cluster_label" in locdata.data.columns


def test_make_Matern():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=10, region=EmptyRegion(), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=1e-10, region=(0, 10), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=1e-10, region=((0, 1), (10, 11)), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=10,
        region=(0, 10),
        cluster_mu=10,
        radius=1.0,
        clip=False,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 12)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=10,
        region=(0, 10),
        cluster_mu=1,
        radius=1.0,
        clip=False,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 12)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=10,
        region=(0, 10),
        cluster_mu=10,
        radius=np.linspace(1, 2, 200),
        clip=False,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-4 <= samples[:, 0])
    assert np.all(samples[:, 0] < 14)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=10,
        region=(0, 10),
        cluster_mu=10,
        radius=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=10,
        region=((0, 1), (10, 11)),
        cluster_mu=10,
        radius=1,
        clip=False,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    with pytest.raises(NotImplementedError):
        samples, labels, parent_samples, region = make_Matern(
            parent_intensity=10,
            region=((0, 1), (10, 11), (100, 101)),
            cluster_mu=10,
            radius=1,
            clip=False,
            shuffle=True,
            seed=rng,
        )
        assert len(samples) == len(labels)
        assert samples.ndim == 2
        assert samples.shape[1] == 3
        assert np.all(-1 <= samples[:, 0])
        assert np.all(samples[:, 0] < 2)
        assert np.all(9 <= samples[:, 1])
        assert np.all(samples[:, 1] < 12)
        assert np.all(99 <= samples[:, 1])
        assert np.all(samples[:, 1] < 102)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=10,
        region=Interval(0, 1),
        cluster_mu=10,
        radius=1,
        clip=False,
        shuffle=True,
        seed=rng,
    )
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=10,
        region=Rectangle((0, 10), 1, 1, 0),
        cluster_mu=10,
        radius=1,
        clip=False,
        shuffle=True,
        seed=rng,
    )
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        cluster_mu=10,
        radius=1,
        clip=False,
        shuffle=True,
        seed=rng,
    )
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=100,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        cluster_mu=10,
        radius=1,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)


@pytest.mark.visual
def test_make_Matern_visual():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=1,
        region=(0, 10),
        cluster_mu=10,
        radius=1.0,
        clip=False,
        shuffle=False,
        seed=rng,
    )

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples, [0] * len(samples), alpha=0.1)
    ax.axis("equal")
    plt.show()

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=0.1,
        region=Polygon(((0, 10), (10, 20), (10, 10))),
        cluster_mu=200,
        radius=np.linspace(0.1, 3, 100),
        clip=False,
        shuffle=False,
        seed=rng,
    )
    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis("equal")
    plt.show()

    samples, labels, parent_samples, region = make_Matern(
        parent_intensity=0.1,
        region=Polygon(((0, 10), (10, 20), (10, 10))),
        cluster_mu=200,
        radius=1,
        clip=True,
        shuffle=True,
        seed=rng,
    )

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis("equal")
    plt.show()


def test_simulate_Matern():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_Matern(
        parent_intensity=10,
        region=(0, 10),
        cluster_mu=10,
        radius=1.0,
        clip=False,
        shuffle=False,
        seed=rng,
    )
    assert locdata.dimension == 1
    assert repr(locdata.region) == "Interval(-1.0, 11.0)"
    assert "cluster_label" in locdata.data.columns


def test_make_Thomas():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10, region=EmptyRegion(), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=0, region=(0, 10), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10, region=(0, 10), cluster_mu=0, seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=1e-10, region=(0, 10), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=(0, 10),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=(0, 10),
        expansion_factor=0,
        cluster_mu=0.1,
        cluster_std=1.0,
        clip=True,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=np.float64(10),
        region=(0, 10),
        expansion_factor=np.float64(1),
        cluster_mu=np.float64(10),
        cluster_std=np.float64(1),
        clip=True,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) > 0

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=(0, 10),
        expansion_factor=1,
        cluster_mu=np.linspace(1, 10, 1000),
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    with pytest.raises(ValueError):
        samples, labels, parent_samples, region = make_Thomas(
            parent_intensity=10,
            region=(0, 10),
            expansion_factor=1,
            cluster_mu=np.linspace(1, 10, 10),
            cluster_std=1.0,
            clip=True,
            shuffle=True,
            seed=rng,
        )

    with pytest.raises(TypeError):
        samples, labels, parent_samples, region = make_Thomas(
            parent_intensity=10,
            region=(0, 10),
            expansion_factor=6,
            cluster_mu=10,
            cluster_std=(1, 2),
            clip=True,
            shuffle=True,
            seed=rng,
        )

    with pytest.raises(TypeError):
        samples, labels, parent_samples, region = make_Thomas(
            parent_intensity=10,
            region=(0, 10),
            expansion_factor=6,
            cluster_mu=10,
            cluster_std=np.linspace(0.2, 2, 200),
            clip=True,
            shuffle=True,
            seed=rng,
        )

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=((0, 1), (10, 11)),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=((0, 1), (10, 11), (100, 101)),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 3
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(94 <= parent_samples[:, 2])
    assert np.all(parent_samples[:, 2] < 107)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=Interval(0, 1),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=Rectangle((0, 10), 1, 1, 0),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=0,
        cluster_mu=np.linspace(10, 20, 200),
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=0,
        cluster_mu=10,
        cluster_std=(0.1, 1),
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 1)
    assert np.all(10 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 11)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=0,
        cluster_mu=10,
        cluster_std=np.linspace((0.01, 0.02), (0.1, 0.2), 100),
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 1)
    assert np.all(10 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 11)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)


@pytest.mark.visual
def test_make_Thomas_visual():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=0.2,
        region=(0, 10),
        expansion_factor=6,
        cluster_mu=100,
        cluster_std=1.0,
        clip=True,
        shuffle=False,
        seed=rng,
    )

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples, [0] * len(samples), alpha=0.1)
    plt.scatter(parent_samples, [0] * len(parent_samples), c="Red", alpha=0.2)
    ax.axis("equal")
    plt.show()

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=50,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=0,
        cluster_mu=100,
        cluster_std=np.linspace((0.001, 0.002), (0.02, 0.04), 100),
        clip=True,
        shuffle=False,
        seed=rng,
    )
    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    plt.scatter(parent_samples[:, 0], parent_samples[:, 1], c="Red")
    ax.axis("equal")
    plt.show()

    samples, labels, parent_samples, region = make_Thomas(
        parent_intensity=20,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=2,
        cluster_mu=500,
        cluster_std=0.02,
        clip=True,
        shuffle=True,
        seed=rng,
    )

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    plt.scatter(parent_samples[:, 0], parent_samples[:, 1], c="Red")
    ax.axis("equal")
    plt.show()


def test_simulate_Thomas():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_Thomas(
        parent_intensity=10,
        region=(0, 10),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=False,
        seed=rng,
    )
    assert locdata.dimension == 1
    assert repr(locdata.region) == "Interval(0, 10)"
    assert "cluster_label" in locdata.data.columns


def test_make_dstorm():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10, region=EmptyRegion(), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=0, region=(0, 10), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=1e-10, region=(0, 10), seed=rng
    )
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=(0, 10),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=np.float64(10),
        region=(0, 10),
        expansion_factor=np.float64(1),
        cluster_mu=np.float64(10),
        cluster_std=np.float64(1),
        clip=True,
        shuffle=False,
        seed=rng,
    )
    assert len(samples) > 0

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=(0, 10),
        expansion_factor=1,
        cluster_mu=np.linspace(1, 10, 1000),
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    with pytest.raises(ValueError):
        samples, labels, parent_samples, region = make_dstorm(
            parent_intensity=10,
            region=(0, 10),
            expansion_factor=1,
            cluster_mu=np.linspace(1, 10, 10),
            cluster_std=1.0,
            clip=True,
            shuffle=True,
            seed=rng,
        )

    with pytest.raises(TypeError):
        samples, labels, parent_samples, region = make_dstorm(
            parent_intensity=10,
            region=(0, 10),
            expansion_factor=6,
            cluster_mu=10,
            cluster_std=(1, 2),
            clip=True,
            shuffle=True,
            seed=rng,
        )

    with pytest.raises(TypeError):
        samples, labels, parent_samples, region = make_dstorm(
            parent_intensity=10,
            region=(0, 10),
            expansion_factor=6,
            cluster_mu=10,
            cluster_std=np.linspace(0.2, 2, 200),
            clip=True,
            shuffle=True,
            seed=rng,
        )

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=((0, 1), (10, 11)),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=((0, 1), (10, 11), (100, 101)),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 3
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(94 <= parent_samples[:, 2])
    assert np.all(parent_samples[:, 2] < 107)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=Interval(0, 1),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=Rectangle((0, 10), 1, 1, 0),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=0,
        cluster_mu=np.linspace(10, 20, 200),
        cluster_std=1.0,
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=0,
        cluster_mu=10,
        cluster_std=(0.1, 1),
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 1)
    assert np.all(10 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 11)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=Polygon(((0, 10), (1, 11), (1, 10))),
        expansion_factor=0,
        cluster_mu=10,
        cluster_std=np.linspace((0.01, 0.02), (0.1, 0.2), 100),
        clip=True,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 1)
    assert np.all(10 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 11)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_dstorm(
        parent_intensity=10,
        region=(0, 10),
        expansion_factor=0,
        cluster_mu=10,
        min_points=5,
        cluster_std=0.1,
        clip=False,
        shuffle=True,
        seed=rng,
    )
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.bincount(labels).mean() == pytest.approx(10, rel=0.2)


def test_simulate_dstorm():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_dstorm(
        parent_intensity=10,
        region=(0, 10),
        expansion_factor=6,
        cluster_mu=10,
        cluster_std=1.0,
        clip=True,
        shuffle=False,
        seed=rng,
    )
    assert locdata.dimension == 1
    assert repr(locdata.region) == "Interval(0, 10)"
    assert "cluster_label" in locdata.data.columns


@pytest.fixture()
def locdata_simple():
    localization_dict = {
        "position_x": [0, 0, 1, 4, 5],
        "position_y": [0, 1, 3, 4, 1],
        "position_z": [0, 1, 3, 4, 1],
        "intensity": [0, 1, 3, 4, 1],
        "uncertainty_x": [10, 30, 100, 300, 0],
        "uncertainty_y": [10, 30, 100, 300, 10],
        "uncertainty_z": [10, 30, 100, 300, 10],
    }
    return LocData(dataframe=pd.DataFrame.from_dict(localization_dict))


def test_resample(locdata_simple):
    dat = resample(locdata=locdata_simple, n_samples=3)
    # print(dat.data)
    assert len(dat) == 15


def test_simulate_tracks():
    dat = simulate_tracks(n_walks=2, n_steps=3)
    # print(dat.data)
    # print(dat.meta)
    assert len(dat) == 6
    assert len(dat.coordinate_labels) == 2


def test__random_walk_drift():
    cumsteps = _random_walk_drift(
        n_steps=10, diffusion_constant=(1, 10), velocity=(0, 0), seed=1
    )
    assert cumsteps.shape == (2, 10)
    cumsteps = _random_walk_drift(
        n_steps=10, diffusion_constant=(0, 0), velocity=(1, 10), seed=1
    )
    assert cumsteps.shape == (2, 10)


def test__drift():
    frames = np.arange(2, 10, 2)
    n_frames = len(frames)
    position_deltas = _drift(frames, diffusion_constant=(1, 10), velocity=None, seed=1)
    assert position_deltas.shape == (2, n_frames)
    position_deltas = _drift(frames, diffusion_constant=None, velocity=(1, 2), seed=1)
    assert position_deltas.shape == (2, n_frames)
    position_deltas = _drift(frames, diffusion_constant=None, velocity=None, seed=1)
    assert position_deltas is None


def test_simulate_drift(locdata_2d):
    new_locdata = add_drift(locdata_2d, diffusion_constant=None, velocity=None, seed=1)
    assert len(new_locdata) == len(locdata_2d)
    new_locdata = add_drift(
        locdata_2d, diffusion_constant=(1, 10), velocity=(10, 10), seed=1
    )
    assert len(new_locdata) == len(locdata_2d)
    # print(new_locdata.meta)


@pytest.mark.visual
def test_visual__drift():
    frames = np.arange(0, 1_000_000, dtype=int)
    print(frames.shape)

    # position_deltas = _drift(frames=frames, velocity=(1, 2), seed=1)
    position_deltas = _drift(frames=frames, diffusion_constant=(1, 2), seed=1)
    print(position_deltas.shape)
    print(position_deltas[0, :10])
    for pd_ in position_deltas:
        plt.plot(frames, pd_)
        plt.plot(frames, pd_)
    plt.show()


@pytest.mark.visual
def test_visual_add_drift(locdata_2d):
    # new_locdata = add_drift(locdata_2d, diffusion_constant=(1, 10), velocity=(10, 10), seed=1)
    new_locdata = add_drift(
        locdata_2d, diffusion_constant=None, velocity=(1, 1), seed=1
    )
    ax = locdata_2d.data.plot(*locdata_2d.coordinate_labels, kind="scatter")
    new_locdata.data.plot(*new_locdata.coordinate_labels, kind="scatter", ax=ax, c="r")
    plt.show()


def test_simulate_frame_numbers(locdata_2d):
    frames = simulate_frame_numbers(n_samples=(len(locdata_2d)), lam=2)
    assert len(frames) == len(locdata_2d)



import tempfile
from pathlib import Path

import matplotlib.pyplot as plt
import pytest

from locan import ROOT_DIR, Ellipse, LocData, Polygon, Rectangle, Roi, rasterize
from locan.data import metadata_pb2


def test_Roi_2d(locdata_2d):
    roi = Roi(reference=locdata_2d, region=Rectangle((0, 0), 2, 2, 0))
    assert (
        repr(roi) == f"Roi("
        f"reference={locdata_2d}, "
        f"region=Rectangle((0, 0), 2, 2, 0),  "
        f"loc_properties=())"
    )
    assert isinstance(roi._region, Rectangle)
    assert isinstance(roi.region, Rectangle)

    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is roi.region

    roi.region = Ellipse((0, 0), 2, 2, 0)
    assert isinstance(roi.region, Ellipse)

    roi = Roi(
        reference=locdata_2d,
        region=Rectangle((0, 0), 4, 4, 0),
        loc_properties=("position_y", "frame"),
    )
    new_dat = roi.locdata()
    assert len(new_dat) == 2
    assert new_dat.region is None

    locdata_empty = LocData()
    roi_empty = Roi(reference=locdata_empty, region=Rectangle((0, 0), 4, 4, 0))
    empty_dat = roi_empty.locdata()
    assert len(empty_dat) == 0
    assert new_dat.region is None

    roi = Roi(
        reference=dict(file_path="test_path", file_type=0),
        region=Rectangle((0, 0), 2, 2, 0),
    )
    # print(roi.reference)
    assert isinstance(roi.reference, metadata_pb2.Metadata)
    assert roi.reference.file.type == 0
    assert roi.reference.file.path == "test_path"

    meta = metadata_pb2.Metadata()
    meta.file.path = "test_path"
    meta.file.type = 0
    roi = Roi(reference=meta, region=Rectangle((0, 0), 2, 2, 0))
    # print(roi.reference)
    assert isinstance(roi.reference, metadata_pb2.Metadata)
    assert roi.reference.file.type == 0
    assert roi.reference.file.path == "test_path"

    meta = metadata_pb2.File()
    meta.path = "test_path"
    meta.type = 0
    roi = Roi(reference=meta, region=Rectangle((0, 0), 2, 2, 0))
    # print(roi.reference)
    assert isinstance(roi.reference, metadata_pb2.Metadata)
    assert roi.reference.file.type == 0
    assert roi.reference.file.path == "test_path"


def test_Roi_3d(locdata_3d):
    roi = Roi(reference=locdata_3d, region=Rectangle((0, 0), 2, 2, 0))
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is None
    assert isinstance(roi.region, Rectangle)
    roi.region = Ellipse((0, 0), 2, 2, 0)
    assert isinstance(roi.region, Ellipse)

    roi = Roi(
        reference=locdata_3d,
        region=Rectangle((0, 0), 2, 3, 0),
        loc_properties=("position_y", "position_z"),
    )
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is None


def test_pickling_locdata_from_Roi(locdata_2d):
    import pickle

    roi = Roi(reference=locdata_2d, region=Rectangle((0, 0), 2, 1, 0))
    new_dat = roi.locdata()

    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / "pickled_locdata.pickle"
        with open(file_path, "wb") as file:
            pickle.dump(new_dat, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, "rb") as file:
            locdata = pickle.load(file)
        assert len(new_dat) == len(locdata)
        assert isinstance(locdata.meta, metadata_pb2.Metadata)


def test_Roi_io(locdata_2d):
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / "roi.yaml"
        # file_path = ROOT_DIR / 'tests/test_data/roi.yaml'

        roi = Roi(reference=locdata_2d, region=Rectangle((0, 0), 2, 1, 10))
        with pytest.warns(UserWarning):
            roi.to_yaml(path=file_path)

        roi = Roi(region=Rectangle((0, 0), 2, 1, 0))
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new.reference is None
        assert isinstance(roi_new.region, Rectangle)
        assert roi_new.loc_properties == []

        roi = Roi(region=Ellipse((0, 0), 2, 1, 0))
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new.reference is None
        assert isinstance(roi_new.region, Ellipse)
        assert roi_new.loc_properties == []

        roi = Roi(region=Polygon(((0, 0), (0, 1), (1, 0.5), (0.8, 0.2))))
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new.reference is None
        assert isinstance(roi_new.region, Polygon)
        assert roi_new.loc_properties == []

        roi = Roi(
            reference=dict(
                file_path=ROOT_DIR / "tests/test_data/five_blobs.txt", file_type=1
            ),
            region=Rectangle((0, 0), 2, 1, 0),
        )
        assert isinstance(roi.reference, (metadata_pb2.Metadata, Path))
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new

        # test region specs with numpy floats
        roi = Roi(
            reference=dict(file_path=file_path, file_type=1),
            region=Rectangle((0, 0), float(2), float(1), float(10)),
        )
        assert isinstance(roi.reference, metadata_pb2.Metadata)
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new

        meta_ = metadata_pb2.Metadata()
        meta_.file.path = str(file_path)
        meta_.file.type = 1
        locdata_2 = LocData.from_selection(locdata_2d, meta=meta_)
        roi = Roi(reference=locdata_2, region=Rectangle((0, 0), 2, 1, 0))
        assert isinstance(roi.reference.meta, metadata_pb2.Metadata)
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new
        assert isinstance(roi_new.reference, metadata_pb2.Metadata)


def test_roi_locdata_from_file(locdata_blobs_3d):
    roi = Roi(reference=locdata_blobs_3d, region=Rectangle((100, 500), 200, 200, 10))
    dat = roi.locdata()
    assert len(dat) == 10

    roi = Roi(
        reference=locdata_blobs_3d,
        region=Rectangle((100, 3), 200, 2, 0),
        loc_properties=["position_x", "cluster_label"],
    )
    dat = roi.locdata()
    assert len(dat) == 10


def test_as_artist():
    roi_rectangle = Roi(reference=None, region=Rectangle((0, 0), 0.5, 0.5, 10))
    roi_ellipse = Roi(reference=None, region=Ellipse((0.5, 0.5), 0.2, 0.3, 10))
    roi_polygon = Roi(
        reference=None,
        region=Polygon(((0, 0.8), (0.1, 0.9), (0.6, 0.9), (0.7, 0.7), (0, 0.8))),
    )
    fig, ax = plt.subplots()
    ax.add_patch(roi_rectangle._region.as_artist())
    ax.add_patch(roi_ellipse._region.as_artist())
    ax.add_patch(roi_polygon._region.as_artist())
    # plt.show()

    plt.close("all")


# to be deprecated
# @pytest.mark.gui
# def test_select_by_drawing_mpl():
#     dat = load_txt_file(path=ROOT_DIR / 'tests/test_data/five_blobs.txt')
#     select_by_drawing_mpl(dat, region_type='rectangle')
#     select_by_drawing_mpl(dat, region_type='ellipse')


# standard LocData fixtures
def test_rasterize(locdata_2d, locdata_single_localization, locdata_non_standard_index):
    res = rasterize(locdata=locdata_2d, support=None, n_regions=(5, 2))
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == "Rectangle((1.0, 1.0), 0.8, 2.5, 0)"
    assert len(res) == 10

    res = rasterize(locdata=locdata_2d, support=((0, 10), (10, 20)), n_regions=(5, 2))
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == "Rectangle((0.0, 10.0), 2.0, 5.0, 0)"
    assert len(res) == 10

    res = rasterize(
        locdata=locdata_single_localization,
        support=((0, 10), (10, 20)),
        n_regions=(5, 2),
    )
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == "Rectangle((0.0, 10.0), 2.0, 5.0, 0)"
    assert len(res) == 10

    res = rasterize(
        locdata=locdata_non_standard_index,
        support=((0, 10), (10, 20)),
        n_regions=(5, 2),
    )
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == "Rectangle((0.0, 10.0), 2.0, 5.0, 0)"
    assert len(res) == 10


def test_rasterize_(locdata_empty):
    with pytest.raises(ValueError):
        rasterize(locdata=locdata_empty, support=None, n_regions=(5, 2))
    with pytest.raises(ValueError):
        rasterize(locdata=locdata_empty, support=((0, 10), (10, 20)), n_regions=(5, 2))


def test_rasterize_3d(locdata_3d):
    res = rasterize(
        locdata=locdata_3d,
        support=None,
        n_regions=(5, 2),
        loc_properties=["position_x", "position_z"],
    )
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == "Rectangle((1.0, 1.0), 0.8, 2.0, 0)"
    assert res[0].loc_properties == ["position_x", "position_z"]
    assert len(res) == 10



import tempfile
from copy import deepcopy
from pathlib import Path

import numpy as np
import pytest
from pandas.testing import assert_frame_equal

import locan as lc
import locan.constants
from locan import load_SMAP_file, load_SMAP_header, load_txt_file, save_SMAP_csv
from locan.dependencies import HAS_DEPENDENCY

pytestmark = pytest.mark.skipif(not HAS_DEPENDENCY["h5py"], reason="requires h5py.")


def test_load_SMAP_header(caplog):
    columns = load_SMAP_header(
        path=lc.ROOT_DIR / "tests/test_data/smap_dstorm_data.mat"
    )
    assert np.array_equal(
        columns,
        [
            "LLrel",
            "PSFxnm",
            "PSFxpix",
            "PSFynm",
            "PSFypix",
            "local_background",
            "channel",
            "filenumber",
            "frame",
            "iterations",
            "locprecnm",
            "locprecznm",
            "logLikelihood",
            "intensity",
            "photerr",
            "position_x",
            "uncertainty_x",
            "xpix",
            "xpixerr",
            "position_y",
            "uncertainty_y",
            "ypix",
            "ypixerr",
            "zerr",
            "position_z",
        ],
    )
    assert caplog.record_tuples[0] == (
        "locan.locan_io.locdata.utilities",
        30,
        "Column LLrel is not a Locan property standard.",
    )


# this is inactive because an empty .mat file is too large to be kept as test_data
# def test_loading_SMAP_file_empty_file():
#     locdata = load_SMAP_file(
#         path=lc.ROOT_DIR / 'tests/test_data/smap_dstorm_data_empty.mat',
#         nrows=10)
#     assert len(locdata) == 0


def test_loading_SMAP_file(caplog):
    locdata = load_SMAP_file(
        path=lc.ROOT_DIR / "tests/test_data/smap_dstorm_data.mat", nrows=10
    )
    assert len(locdata) == 10
    assert np.array_equal(
        locdata.data.columns,
        [
            "LLrel",
            "PSFxnm",
            "PSFxpix",
            "PSFynm",
            "PSFypix",
            "local_background",
            "channel",
            "filenumber",
            "frame",
            "iterations",
            "locprecnm",
            "locprecznm",
            "logLikelihood",
            "intensity",
            "photerr",
            "position_x",
            "uncertainty_x",
            "xpix",
            "xpixerr",
            "position_y",
            "uncertainty_y",
            "ypix",
            "ypixerr",
            "zerr",
            "position_z",
        ],
    )
    assert caplog.record_tuples[0] == (
        "locan.locan_io.locdata.utilities",
        30,
        "Column LLrel is not a Locan property standard.",
    )


def test_save_and_load_smap_csv(locdata_2d):
    locdata = deepcopy(locdata_2d)
    # introduce different dtypes in data
    locdata.data["position_x"] = locdata.data["position_x"].astype("float64")
    locdata.data["position_y"] = locdata.data["position_y"].astype("int32")
    locdata.data["frame"] = locdata.data["frame"].astype("int32")
    locdata.data["intensity"] = locdata.data["intensity"].astype("int32")
    assert all(locdata.data.dtypes.values == ["float64", "int32", "int32", "int32"])
    # print(locdata.data)

    with tempfile.TemporaryDirectory() as tmp_directory:
        # save smap file
        file_path = Path(tmp_directory) / "locdata.csv"
        save_SMAP_csv(locdata, path=file_path)

        # read back smap file
        locdata_ = load_txt_file(
            path=file_path, property_mapping=locan.constants.SMAP_KEYS
        )
        assert len(locdata_) == len(locdata)
        assert locdata_.properties.keys() == locdata.properties.keys()

        # print(locdata_.data)
        # print(locdata_.data.dtypes)
        assert_frame_equal(locdata_.data, locdata.data, check_dtype=False)
        # todo: add convert kwarg to load_txt_file
        # assert_frame_equal(locdata_.data, locdata.data, check_dtype=True)



import matplotlib.pyplot as plt  # this import is needed for visual inspection
import numpy as np
import pytest
from lmfit.model import ModelResult
from lmfit.models import ConstantModel, LinearModel, PolynomialModel

from locan import Drift, DriftComponent, LocData
from locan.analysis.drift import (
    _ConstantModelFacade,
    _ConstantOneModelFacade,
    _ConstantZeroModelFacade,
    _estimate_drift_cc,
    _estimate_drift_icp,
    _LmfitModelFacade,
    _SplineModelFacade,
)
from locan.dependencies import HAS_DEPENDENCY

# data to evaluate fitting
x = np.array([1, 2, 4, 6, 9, 10, 11, 15, 16, 20])
y = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])


def test__LmfitModelFacade():
    model = _LmfitModelFacade(LinearModel())
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == pytest.approx(3.401234567901236)
    assert len(model.eval([2, 5])) == 2
    model.plot()
    # plt.show()

    plt.close("all")


def test__ConstantModelFacade():
    model = _ConstantModelFacade()
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == pytest.approx(5.500000000011998)
    assert len(model.eval([2, 5])) == 2
    model.plot()
    # plt.show()

    plt.close("all")


def test__ConstantZeroModelFacade():
    model = _ConstantZeroModelFacade()
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == 0
    assert len(model.eval([2, 5])) == 2
    model.plot()
    # plt.show()

    plt.close("all")


def test__ConstantOneModelFacade():
    model = _ConstantOneModelFacade()
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == 1
    assert len(model.eval([2, 5])) == 2
    model.plot()
    # plt.show()

    plt.close("all")


def test__SplineModelFacade():
    model = _SplineModelFacade()
    model.fit(x=x, y=y, verbose=True)
    # print(model.eval(5))
    # print(model.eval((3, 5)))
    # print(model.eval([3, 5]))
    # print(model.eval(np.array([3, 5])))
    assert model.eval(5) == pytest.approx(3.417849958061275)
    assert len(model.eval([2, 5])) == 2
    # model.plot()
    model.fit(x=x, y=y, s=0)
    assert model.eval(5) == pytest.approx(3.5332295857553917)
    # model.plot()
    # plt.show()
    model = _SplineModelFacade(s=0)
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == pytest.approx(3.5332295857553917)


def test_DriftComponent():
    drift_component = DriftComponent(type="linear")
    assert drift_component.model_result is None
    with pytest.raises(AttributeError):
        drift_component.eval(3)

    with pytest.raises(TypeError):
        DriftComponent(type="undefined")

    drift_component = DriftComponent(type="linear").fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, LinearModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.eval(4) == pytest.approx(2.924242424242425)
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type="zero").fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, ConstantModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.model_result.best_fit == 0
    assert drift_component.eval(4) == 0
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type="one").fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, ConstantModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.model_result.best_fit == 1
    assert drift_component.eval(4) == 1
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type="constant").fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, ConstantModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.model_result.best_fit == pytest.approx(5.500000000011998)
    assert drift_component.eval(4) == pytest.approx(5.500000000011998)
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type="polynomial", degree=2).fit(
        x, y, verbose=True
    )
    assert isinstance(drift_component.model.model, PolynomialModel)
    assert isinstance(drift_component.model_result, ModelResult)

    drift_component = DriftComponent(type="spline").fit(x, y, verbose=True, s=0.3)
    assert drift_component.eval(4) == pytest.approx(3.0972390587725087)
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type=LinearModel()).fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, LinearModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.eval(4) == pytest.approx(2.924242424242425)
    assert len(drift_component.eval([1, 4]) == 2)
    # plt.show()

    plt.close("all")


def test_Drift_empty(caplog):
    drift = Drift().compute(LocData())
    drift.plot()
    drift.apply_correction()
    drift.fit_transformation()
    drift.fit_transformations()
    assert caplog.record_tuples == [
        ("locan.analysis.drift", 30, "Locdata is empty."),
        ("locan.analysis.drift", 30, "No transformations available to be applied."),
        ("locan.analysis.drift", 30, "No transformations available to be fitted."),
        ("locan.analysis.drift", 30, "No transformations available to be fitted."),
    ]


def test__estimate_drift_cc(locdata_blobs_2d):
    # locan.render_2d_mpl(locdata_blobs_2d, bin_size=10, rescale=locan.Trafo.EQUALIZE)
    # plt.show()

    collection, transformations = _estimate_drift_cc(
        locdata_blobs_2d,
        chunks=None,
        chunk_size=25,
        target="first",
        bin_size=10,
        kwargs_chunk=None,
        kwargs_register=None,
    )
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2

    index_groups = (range(0, 25), range(25, 50))
    collection, transformations = _estimate_drift_cc(
        locdata_blobs_2d,
        chunks=index_groups,
        chunk_size=None,
        target="first",
        bin_size=10,
        kwargs_chunk=None,
        kwargs_register=None,
    )
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2

    collection, transformations = _estimate_drift_cc(
        locdata_blobs_2d,
        chunks=None,
        chunk_size=25,
        target="previous",
        bin_size=10,
        kwargs_chunk=None,
        kwargs_register=None,
    )
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test__estimate_drift_icp(locdata_blobs_2d):
    collection, transformations = _estimate_drift_icp(
        locdata_blobs_2d,
        chunks=None,
        chunk_size=25,
        target="first",
        kwargs_chunk=None,
        kwargs_register=None,
    )
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2

    index_groups = (range(0, 25), range(25, 50))
    collection, transformations = _estimate_drift_icp(
        locdata_blobs_2d,
        chunks=index_groups,
        chunk_size=None,
        target="first",
        kwargs_chunk=None,
        kwargs_register=None,
    )
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2

    collection, transformations = _estimate_drift_icp(
        locdata_blobs_2d,
        chunks=None,
        chunk_size=25,
        target="previous",
        kwargs_chunk=None,
        kwargs_register=None,
    )
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2


def test_Drift(locdata_blobs_2d):
    drift = Drift(chunk_size=15, target="first", method="cc").compute(locdata_blobs_2d)
    assert isinstance(drift.locdata, LocData)
    assert isinstance(drift.collection, LocData)
    assert drift.transformations[0]._fields == ("matrix", "offset")
    assert len(drift.transformations) == 4
    assert drift.transformation_models["matrix"] is None
    assert drift.transformation_models["offset"] is None

    std_trafo_models = drift._transformation_models_for_identity_matrix()
    assert std_trafo_models["matrix"][0].type == "one"
    std_trafo_models = drift._transformation_models_for_zero_offset()
    assert std_trafo_models["offset"][0].type == "zero"

    assert drift.transformation_models["matrix"] is None
    drift.fit_transformation(
        slice_data=slice(None),
        transformation_component="matrix",
        element=1,
        drift_model="linear",
        verbose=False,
    )
    assert isinstance(drift.transformation_models["matrix"][0], DriftComponent)
    assert drift.transformation_models["matrix"][0].type == "one"
    assert drift.transformation_models["matrix"][1].type == "linear"
    assert drift.transformation_models["matrix"][2].type == "zero"

    assert drift.transformation_models["offset"] is None
    drift.fit_transformation(
        slice_data=slice(None),
        transformation_component="offset",
        element=1,
        drift_model="linear",
        verbose=False,
    )
    assert drift.transformation_models["offset"][0].type == "zero"
    assert drift.transformation_models["offset"][1].type == "linear"

    drift.fit_transformation(
        slice_data=slice(None),
        transformation_component="matrix",
        element=1,
        drift_model=None,
        verbose=False,
    )
    assert drift.transformation_models["matrix"][1].type == "linear"

    drift.fit_transformations(matrix_models=None, offset_models=None, verbose=False)
    assert drift.transformation_models["matrix"] is None
    assert drift.transformation_models["offset"] is None

    drift.fit_transformations(
        slice_data=slice(None),
        matrix_models=(
            "constant",
            "linear",
            DriftComponent(type="polynomial", degree=2),
            None,
        ),
        offset_models=("spline", DriftComponent(type="spline", s=100)),
        verbose=False,
    )
    assert drift.transformation_models["matrix"][0].type == "constant"
    assert drift.transformation_models["matrix"][1].type == "linear"
    assert drift.transformation_models["matrix"][2].type == "polynomial"
    assert drift.transformation_models["matrix"][3].type == "one"
    assert drift.transformation_models["offset"][0].type == "spline"
    assert drift.transformation_models["offset"][1].type == "spline"

    assert drift.locdata_corrected is None
    new_locdata = drift._apply_correction_on_chunks()
    assert len(new_locdata) == len(locdata_blobs_2d)

    drift.transformation_models["matrix"] = None
    drift.transformation_models["offset"] = None
    with pytest.raises(AttributeError):
        drift._apply_correction_from_model(drift.locdata_rapidSTORM_2d)

    drift.fit_transformation(
        slice_data=slice(None),
        transformation_component="matrix",
        element=1,
        drift_model="linear",
        verbose=True,
    )
    assert drift.transformation_models["matrix"][1].type == "linear"
    # print(drift.transformation_models['matrix'][1].model)
    # print(drift.transformation_models['matrix'][1].model_result)
    # print(drift.transformation_models['matrix'][1].eval(3))

    drift.fit_transformation(
        slice_data=slice(None),
        transformation_component="matrix",
        element=0,
        drift_model="one",
        verbose=True,
    )
    assert drift.transformation_models["matrix"][0].type == "one"
    # print(drift.transformation_models['matrix'][0].model)
    # print(drift.transformation_models['matrix'][0].model_result)
    # print(drift.transformation_models['matrix'][0].eval(3))

    drift.fit_transformations(
        slice_data=slice(None),
        matrix_models=["linear"] * 4,
        offset_models=["linear"] * 2,
        verbose=False,
    )
    transformed_points = drift._apply_correction_from_model(drift.locdata)
    assert len(transformed_points) == len(locdata_blobs_2d)

    drift.apply_correction(from_model=True)
    assert isinstance(drift.locdata_corrected, LocData)
    drift.apply_correction(from_model=False)
    assert isinstance(drift.locdata_corrected, LocData)
    drift.apply_correction(locdata=drift.collection.references[0], from_model=True)
    assert isinstance(drift.locdata_corrected, LocData)
    assert len(drift.locdata_corrected) == len(drift.collection.references[0])

    # change target and fit_transformations
    drift = Drift(chunk_size=25, target="previous", method="cc").compute(
        locdata_blobs_2d
    )
    assert isinstance(drift.collection, LocData)
    assert drift.transformations[0]._fields == ("matrix", "offset")
    assert len(drift.transformations) == 2

    drift.plot(transformation_component="matrix", element=None)
    drift.plot(transformation_component="matrix", element=3)
    drift.plot(transformation_component="offset", element=None)
    drift.plot(transformation_component="offset", element=0)
    # plt.show()

    with pytest.raises(NotImplementedError):
        drift.fit_transformations(
            slice_data=slice(None),
            matrix_models=["linear"] * 4,
            offset_models=["linear"] * 2,
            verbose=False,
        )

    plt.close("all")


def test_Drift_chain(locdata_blobs_2d):
    # apply chained functions
    drift = (
        Drift(chunk_size=25, target="first", method="cc")
        .compute(locdata_blobs_2d)
        .fit_transformations(
            slice_data=slice(0, 2),
            matrix_models=None,
            offset_models=(None, LinearModel()),
        )
        .apply_correction()
    )
    assert len(drift.locdata_corrected) == len(drift.locdata)
    assert isinstance(drift.locdata, LocData)
    assert isinstance(drift.collection, LocData)
    assert drift.transformations[0]._fields == ("matrix", "offset")
    assert len(drift.transformations) == 2


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_Drift_with_icp(locdata_blobs_2d):
    drift = Drift(chunk_size=15, target="first", method="icp").compute(locdata_blobs_2d)
    assert isinstance(drift.locdata, LocData)
    assert isinstance(drift.collection, LocData)
    assert drift.transformations[0]._fields == ("matrix", "offset")
    assert len(drift.transformations) == 4
    assert drift.transformation_models["matrix"] is None
    assert drift.transformation_models["offset"] is None

    assert drift.transformation_models["matrix"] is None
    drift.fit_transformation(
        slice_data=slice(None),
        transformation_component="matrix",
        element=1,
        drift_model="linear",
        verbose=False,
    )
    assert isinstance(drift.transformation_models["matrix"][0], DriftComponent)
    assert drift.transformation_models["matrix"][0].type == "one"
    assert drift.transformation_models["matrix"][1].type == "linear"
    assert drift.transformation_models["matrix"][2].type == "zero"

    assert drift.transformation_models["offset"] is None
    drift.fit_transformation(
        slice_data=slice(None),
        transformation_component="offset",
        element=1,
        drift_model="linear",
        verbose=False,
    )
    assert drift.transformation_models["offset"][0].type == "zero"
    assert drift.transformation_models["offset"][1].type == "linear"



import matplotlib.pyplot as plt  # noqa: F401  # this import is needed for interactive tests
import numpy as np
import pytest

import locan
from locan import (  # noqa: F401  # this import is needed for interactive tests
    RenderEngine,
    render_2d_mpl,
    render_2d_napari,
    render_2d_rgb_napari,
    transform_affine,
)
from locan.dependencies import HAS_DEPENDENCY

if HAS_DEPENDENCY["napari"]:
    import napari

pytestmark = [
    pytest.mark.qt,
    pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="requires napari"),
]

HAS_NAPARI_AND_PYTESTQT = HAS_DEPENDENCY["napari"] and HAS_DEPENDENCY["pytestqt"]
# pytestqt is not a requested or extra dependency.
# If napari and pytest-qt is installed, all tests run.
# Tests in docker or GitHub actions on linux require xvfb
# for tests with pytest-qt to run.


@pytest.mark.visual
# this is to check overlay of rendered image and single localization points
def test_render_2d_napari_new_visual(locdata_blobs_2d):
    print(locdata_blobs_2d.data[locdata_blobs_2d.coordinate_labels].describe())

    render_2d_mpl(locdata_blobs_2d, bin_size=10, cmap="viridis")
    plt.show()

    viewer = napari.Viewer()

    bins = locan.Bins(
        bin_size=(50, 100),
        bin_range=(
            (0, locdata_blobs_2d.coordinates.max(axis=0)[0]),
            (0, locdata_blobs_2d.coordinates.max(axis=0)[1]),
        ),
        extend_range=True,
    )

    render_2d_napari(locdata_blobs_2d, bins=bins, cmap="viridis", viewer=viewer)

    render_2d_napari(locdata_blobs_2d, bin_size=10, cmap="viridis", viewer=viewer)

    viewer.add_points(locdata_blobs_2d.coordinates, size=10, opacity=0.2)

    print(viewer.layers[0].corner_pixels)
    print(viewer.layers[0].data_to_world(viewer.layers[0].corner_pixels))

    napari.run()
    plt.close("all")


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_2d_napari(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()

    with pytest.raises(AssertionError):
        render_2d_napari(
            locdata_blobs_2d, viewer=viewer, bin_edges=((0, 10, 100), (0, 10, 100))
        )

    viewer_ = render_2d_napari(
        locdata_blobs_2d, viewer=viewer, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert viewer_ is viewer
    assert len(viewer.layers) == 1
    assert viewer.layers[0].name == "LocData 0"
    assert np.array_equal(viewer.layers[0].corner_pixels, [[0, 0], [8, 5]])
    assert np.array_equal(
        viewer.layers[0].data_to_world(viewer.layers[0].corner_pixels),
        [[112.0, 465.0], [912.0, 965.0]],
    )
    assert viewer.scale_bar.unit is None or len(viewer.scale_bar.unit) != 0
    assert viewer.layers[0].metadata["message"]
    viewer.close()


@pytest.mark.gui
def test_render_2d_napari_empty_gui(locdata_empty):
    render_2d_napari(locdata_empty, bin_size=100, cmap="viridis", gamma=0.1)
    napari.run()


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_2d_napari_empty(make_napari_viewer, locdata_empty):
    viewer = make_napari_viewer()
    render_2d_napari(
        locdata_empty, viewer=viewer, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert viewer.layers == []
    viewer.close()


@pytest.mark.gui
def test_render_2d_napari_single_gui(locdata_single_localization, caplog):
    render_2d_napari(
        locdata_single_localization, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert caplog.record_tuples[1] == (
        "locan.render.render2d",
        30,
        "Locdata carries a single localization.",
    )
    napari.run()


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_2d_napari_single(
    make_napari_viewer, locdata_single_localization, caplog
):
    viewer = make_napari_viewer()
    render_2d_napari(
        locdata_single_localization,
        viewer=viewer,
        bin_size=100,
        cmap="viridis",
        gamma=0.1,
    )
    viewer.close()
    assert caplog.record_tuples[0] == (
        "locan.visualize.napari.render2d",
        30,
        "Locdata carries a single localization.",
    )


@pytest.mark.gui
def test_render_2d_rgb_napari_gui(locdata_blobs_2d):
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))
    render_2d_rgb_napari([locdata_0, locdata_1], bin_size=20)

    napari.run()


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_2d_rgb_napari(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))
    render_2d_rgb_napari([locdata_0, locdata_1], viewer=viewer, bin_size=20)
    viewer.close()



import numpy as np
import pytest

from locan import Roi, get_rois, save_rois, select_by_drawing_napari
from locan.dependencies import HAS_DEPENDENCY
from locan.visualize.napari.utilities import _shape_to_region, _shapes_to_regions

if HAS_DEPENDENCY["napari"]:
    import napari

pytestmark = [
    pytest.mark.qt,
    pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="requires napari"),
]

HAS_NAPARI_AND_PYTESTQT = HAS_DEPENDENCY["napari"] and HAS_DEPENDENCY["pytestqt"]
# pytestqt is not a requested or extra dependency.
# If napari and pytest-qt is installed, all tests run.
# Tests in docker or GitHub actions on linux require xvfb
# for tests with pytest-qt to run.


def clean_directory(directory):
    for item in directory.glob("*"):
        if item.is_dir():
            item.rmdir()
        else:
            item.unlink()


def test__shape_to_region():
    # rectangle
    vertices = np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]])
    region = _shape_to_region(vertices, "rectangle")
    assert repr(region) == "Rectangle((0.0, 0.0), 3.1, 2.5, 0)"

    # ellipse
    vertices = np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]])
    region = _shape_to_region(vertices, "ellipse")
    assert repr(region) == "Ellipse((1.55, 1.25), 3.1, 2.5, 0)"

    # polygon
    vertices = np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]])
    region = _shape_to_region(vertices, "polygon")
    assert (
        repr(region)
        == "Polygon([[0.0, 0.0], [0.0, 2.5], [3.1, 2.5], [3.1, 0.0], [0.0, 0.0]])"
    )


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test__shapes_to_regions(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()
    shape_data = (np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]]), "rectangle")
    layer_shapes = viewer.add_shapes(shape_data)

    shape_data = (np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]]), "ellipse")
    layer_shapes.add(shape_data)

    shape_data = (np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]]), "polygon")
    layer_shapes.add(shape_data)

    shapes_data = layer_shapes.as_layer_data_tuple()
    regions = _shapes_to_regions(shapes_data)

    expected_regions = [
        "Rectangle((0.0, 0.0), 3.1, 2.5, 0)",
        "Ellipse((1.55, 1.25), 3.1, 2.5, 0)",
        "Polygon([[0.0, 0.0], [0.0, 2.5], [3.1, 2.5], [3.1, 0.0], [0.0, 0.0]])",
    ]

    for region, expected in zip(regions, expected_regions):
        assert repr(region) == expected


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_get_rois(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()
    shape_data = (np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]]), "rectangle")
    layer_shapes = viewer.add_shapes(shape_data)

    shape_data = (np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]]), "ellipse")
    layer_shapes.add(shape_data)

    shape_data = (np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]]), "polygon")
    layer_shapes.add(shape_data)

    rois = get_rois(
        shapes_layer=layer_shapes,
        reference=locdata_blobs_2d,
        loc_properties=locdata_blobs_2d.coordinate_labels,
    )
    for roi in rois:
        assert isinstance(roi, Roi)
        assert roi.loc_properties == ["position_x", "position_y"]


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_save_rois(make_napari_viewer, locdata_blobs_2d, tmp_path):
    viewer = make_napari_viewer()
    shape_data = (np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]]), "rectangle")
    layer_shapes = viewer.add_shapes(shape_data)

    shape_data = (np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]]), "ellipse")
    layer_shapes.add(shape_data)

    shape_data = (np.array([[0, 0], [0, 2.5], [3.1, 2.5], [3.1, 0]]), "polygon")
    layer_shapes.add(shape_data)

    rois = get_rois(
        shapes_layer=layer_shapes,
        reference=locdata_blobs_2d,
        loc_properties=locdata_blobs_2d.coordinate_labels,
    )

    # files_path is existing directory
    roi_path_list = save_rois(rois=rois, file_path=tmp_path)
    saved_path_list = list(tmp_path.glob("*.yaml"))
    for roi_path, saved_path in zip(roi_path_list, saved_path_list):
        assert roi_path == saved_path
        assert saved_path.suffix == ".yaml"
        assert saved_path.stem[:-2] == "my_roi"
    clean_directory(tmp_path)

    # file_path is in non-existing directory
    with pytest.raises(FileNotFoundError):
        save_rois(rois=rois, file_path=tmp_path / "name" / "name.txt")
    clean_directory(tmp_path)

    # file_path is some name in existing directory
    roi_path_list = save_rois(rois=rois, file_path=tmp_path / "some.txt")
    saved_path_list = list(tmp_path.glob("*.yaml"))
    for roi_path, saved_path in zip(roi_path_list, saved_path_list):
        assert roi_path == saved_path
        assert saved_path.suffix == ".yaml"
        assert saved_path.stem[:-2] == "some_roi"
    clean_directory(tmp_path)

    # file_path is roi_reference
    with pytest.raises(AttributeError):
        save_rois(rois=rois, file_path="roi_reference")
    clean_directory(tmp_path)

    # file_path is roi_reference
    modified_rois = []
    for roi in rois:
        roi_ = Roi(
            region=roi.region,
            reference=dict(file_path=tmp_path / "reference.txt", file_type=0),
            loc_properties=roi.loc_properties,
        )
        modified_rois.append(roi_)

    roi_path_list = save_rois(rois=modified_rois, file_path="roi_reference")
    saved_path_list = list(tmp_path.glob("*.yaml"))
    for roi_path, saved_path in zip(roi_path_list[:1], saved_path_list):
        assert roi_path == saved_path
        assert saved_path.suffix == ".yaml"
        assert saved_path.stem[:-2] == "reference_roi"
    clean_directory(tmp_path)

    # todo: add test for file_path=None which seems to crash the file dialog
    #  when napari is running.


@pytest.mark.gui
def test_select_by_drawing_napari_simple_check(locdata_blobs_2d):
    rois = select_by_drawing_napari(locdata_blobs_2d, bin_size=50)
    print(rois)
    assert len(rois) != 0


@pytest.mark.gui
def test_select_by_drawing_napari_gui(locdata_blobs_2d):
    viewer = napari.Viewer()
    viewer.add_shapes(
        name="Rois", data=np.array([(100, 500), (600, 700)]), shape_type="rectangle"
    )

    rois = select_by_drawing_napari(
        locdata_blobs_2d, viewer=viewer, bin_size=10, cmap="viridis", gamma=0.1
    )
    # No need for napari.run() since it is called inside select_by_drawing_napari.
    print(rois)
    print(viewer.layers["Rois"].data)
    assert len(rois) == 1
    assert repr(rois[0].region) == "Rectangle((100.0, 500.0), 500.0, 200.0, 0)"


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_select_by_drawing_napari(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()
    viewer.add_shapes(
        name="Rois", data=np.array([(100, 500), (600, 700)]), shape_type="rectangle"
    )

    rois = select_by_drawing_napari(
        locdata_blobs_2d,
        viewer=viewer,
        napari_run=False,
        bin_size=10,
        cmap="viridis",
        gamma=0.1,
    )
    assert len(rois) == 1
    assert repr(rois[0].region) == "Rectangle((100.0, 500.0), 500.0, 200.0, 0)"
    viewer.close()



"""

Functions for user interaction with paths and file names.

"""
from locan.dependencies import HAS_DEPENDENCY, needs_package

if HAS_DEPENDENCY["qt"]:
    from qtpy.QtWidgets import QApplication, QFileDialog

from locan.configuration import QT_BINDING
from locan.dependencies import QtBindings

__all__ = ["file_dialog"]


@needs_package("qt")
def file_dialog(
    directory=None,
    message="Select a file...",
    filter="Text files (*.txt);; All files (*)",
):
    """
    Select file names in a ui dialog.

    Parameters
    ----------
    directory : str or None
        directory path to start dialog in. If None the current directory is used.

    message : str
        Hint what to do

    filter : str
        filter for file type

    Returns
    -------
    list of str
        list with file names or empty list
    """
    if directory is None:
        directory_ = "./"
    else:
        directory_ = str(directory)

    if QT_BINDING == QtBindings.PYSIDE2:
        app = (
            QApplication.instance()
        )  # this is needed if the function is called twice in a row.
        if app is None:
            app = QApplication([])  # todo: [directory_] is not working - please fix!
    else:
        app = QApplication([directory_])

    fname = QFileDialog.getOpenFileNames(
        None,
        message,
        directory_,
        filter=filter
        # kwargs: parent, message, directory, filter
        # but kw_names are different for different qt_bindings
    )

    if isinstance(fname, tuple):
        return fname[0]
    else:
        return str(fname)



"""

Methods for clustering localization data in LocData objects.

"""
from __future__ import annotations

import sys
from copy import copy

import numpy as np
from sklearn.cluster import DBSCAN

from locan.configuration import N_JOBS
from locan.data.aggregate import Bins, _accumulate_2d, _check_loc_properties, ranges
from locan.data.locdata import LocData
from locan.dependencies import HAS_DEPENDENCY, needs_package

if HAS_DEPENDENCY["hdbscan"]:
    from hdbscan import HDBSCAN

__all__ = ["cluster_hdbscan", "cluster_dbscan", "cluster_by_bin"]


@needs_package("hdbscan")
def cluster_hdbscan(
    locdata,
    min_cluster_size=5,
    loc_properties=None,
    allow_single_cluster=False,
    **kwargs,
):
    """
    Cluster localizations in locdata using the hdbscan clustering algorithm.

    Parameters
    ----------
    locdata : LocData
        Localization data on which to perform the manipulation.
    loc_properties : list[str] | None
        The LocData properties to be used for clustering.
        If None, `locdata.coordinates` will be used.
    min_cluster_size : int
        Minimumm cluster size in HDBSCAN algorithm (default: 5)
    allow_single_cluster : bool
        If True, return single cluster (default: False)
    kwargs : dict
        Other parameters passed to `hdbscan.HDBSCAN`.

    Returns
    -------
    tuple[LocData, LocData]
        A tuple with noise and cluster.
        The first LocData object is a selection of all localizations that are
        defined as noise, in other words all localizations that are not part
        of any cluster.
        The second LocData object is a LocData instance assembling all
        generated selections (i.e. localization cluster).
    """
    parameter = locals()

    if len(locdata) == 0:
        locdata_noise = LocData()
        collection = LocData()

    if len(locdata) < min_cluster_size:
        locdata_noise = copy(locdata)
        collection = LocData()

    else:
        if loc_properties is None:
            fit_data = locdata.coordinates
        else:
            fit_data = locdata.data[loc_properties]

        labels = HDBSCAN(
            min_cluster_size=min_cluster_size,
            allow_single_cluster=allow_single_cluster,
            gen_min_span_tree=False,
            **kwargs,
        ).fit_predict(fit_data)

        grouped = locdata.data.groupby(labels)
        locdata_index_labels = [
            locdata.data.index[idxs] for idxs in grouped.indices.values()
        ]
        selections = [
            LocData.from_selection(locdata=locdata, indices=idxs)
            for idxs in locdata_index_labels
        ]

        try:
            grouped.get_group(-1)
            locdata_noise = selections[0]
            collection = LocData.from_collection(selections[1:])
        except KeyError:
            locdata_noise = None
            collection = LocData.from_collection(selections)

    # set regions
    if locdata_noise:
        locdata_noise.region = locdata.region
    if collection:
        collection.region = locdata.region

    # metadata
    if locdata_noise:
        del locdata_noise.meta.history[:]
        locdata_noise.meta.history.add(
            name=sys._getframe().f_code.co_name, parameter=str(parameter)
        )
    del collection.meta.history[:]
    collection.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(parameter)
    )

    return locdata_noise, collection


def cluster_dbscan(locdata, eps=20, min_samples=5, loc_properties=None, **kwargs):
    """
    Cluster localizations in locdata using the dbscan clustering algorithm as
    implemented in sklearn.

    Parameters
    ----------
    locdata : LocData
        specifying the localization data on which to perform the manipulation.
    eps : float
        The maximum distance between two samples for them to be considered as
        in the same neighborhood.
    min_samples : int
        The number of samples in a neighborhood for a point to be considered
        as a core point.
        This includes the point itself.
    loc_properties : list[str] | None
        The LocData properties to be used for clustering. If None,
        `locdata.coordinates` will be used.
    kwargs : dict
        Other parameters passed to `sklearn.cluster.DBSCAN`.

    Returns
    -------
    tuple[LocData, LocData]
        A tuple with noise and cluster.
        The first LocData object is a selection of all localizations that are
        defined as noise, in other words all localizations that are not part
        of any cluster.
        The second LocData object is a LocData instance assembling all
        generated selections (i.e. localization cluster).
    """
    parameter = locals()

    if len(locdata) == 0:
        locdata_noise = LocData()
        collection = LocData()

    else:
        if loc_properties is None:
            fit_data = locdata.coordinates
        else:
            fit_data = locdata.data[loc_properties]

        labels = DBSCAN(
            eps=eps, min_samples=min_samples, n_jobs=N_JOBS, **kwargs
        ).fit_predict(fit_data)

        grouped = locdata.data.groupby(labels)
        locdata_index_labels = [
            locdata.data.index[idxs] for idxs in grouped.indices.values()
        ]
        selections = [
            LocData.from_selection(locdata=locdata, indices=idxs)
            for idxs in locdata_index_labels
        ]

        try:
            grouped.get_group(-1)
            locdata_noise = selections[0]
            collection = LocData.from_collection(selections[1:])
        except KeyError:
            locdata_noise = None
            collection = LocData.from_collection(selections)

    # set regions
    if locdata_noise:
        locdata_noise.region = locdata.region
    if collection:
        collection.region = locdata.region

    # metadata
    if locdata_noise:
        del locdata_noise.meta.history[:]
        locdata_noise.meta.history.add(
            name=sys._getframe().f_code.co_name, parameter=str(parameter)
        )
    del collection.meta.history[:]
    collection.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(parameter)
    )

    return locdata_noise, collection


def cluster_by_bin(
    locdata,
    loc_properties=None,
    min_samples=1,
    bins=None,
    n_bins=None,
    bin_size=None,
    bin_edges=None,
    bin_range=None,
    return_counts=False,
):
    """
    Cluster localizations in locdata by binning all localizations with regard
    to `loc_properties` and collecting all localizations per bin as cluster.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    loc_properties : list[str] | None
        Localization properties to be grouped into bins.
        If None The coordinate_values of locdata are used.
    min_samples : int
        The minimum number of samples per bin to be considered as cluster.
    bins : int | sequence | Bins | boost_histogram.axis.Axis | None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple | list | numpy.ndarray[float] | None
        Array of bin edges with shape (n_bin_edges,)
        or (dimension, n_bin_edges) for all or each dimension.
    n_bins : int | list[int] | tuple[int] | numpy.ndarray[int] | None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float | list[float] | tuple[float] | numpy.ndarray[float] | None
        The size of bins in units of locdata coordinate units for all or each
        dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple[float] | tuple[tuple[float]] | str | None
        The data bin_range to be taken into consideration for all or each
        dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    return_counts : bool
        If true, n_elements per bin are returned.

    Returns
    -------
    tuple[Bins, numpy.ndarray, LocData, numpy.ndarray | None]
        Tuple with bins, bin_indices,
        collection of all generated selections (i.e. localization clusters),
        and counts per bin.
    """
    parameter = locals()

    if len(locdata) == 0:
        bins = None
        bin_indices = np.array([])
        collection = LocData()
        counts = np.array([]) if return_counts else None
        return bins, bin_indices, collection, counts

    loc_properties = _check_loc_properties(locdata, loc_properties)
    data = locdata.data[loc_properties].values
    if (bin_range is None or isinstance(bin_range, str)) and bin_edges is None:
        bin_range_ = ranges(locdata, loc_properties=loc_properties, special=bin_range)
    else:
        bin_range_ = bin_range

    try:
        bins = Bins(
            bins, n_bins, bin_size, bin_edges, bin_range_, labels=loc_properties
        )
    except ValueError:  # this is raised again to add accustomed message
        raise ValueError("Bin dimension and len of `loc_properties` is incompatible.")

    bin_indices, data_indices, _, counts = _accumulate_2d(
        data, bin_edges=bins.bin_edges, return_counts=True
    )

    if min_samples > 1:
        mask = counts >= min_samples
        counts = counts[mask]
        bin_indices = bin_indices[mask]
        data_indices = [
            data_indices_e
            for data_indices_e, mask_e in zip(data_indices, mask)
            if mask_e
        ]

    selections = [
        LocData.from_selection(locdata=locdata, indices=idxs) for idxs in data_indices
    ]
    collection = LocData.from_collection(selections)

    # set regions
    if collection:
        collection.region = locdata.region

    # metadata
    del collection.meta.history[:]
    collection.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(parameter)
    )

    if not return_counts:
        counts = None

    return bins, bin_indices, collection, counts



"""
Benchmark functions to be used with Airspeed Velocity.
"""
import numpy as np
import pandas as pd

import locan as lc

rng = np.random.default_rng(seed=1)


class BenchmarkMetadata:
    """
    Benchmarks for data.metadata_pb2.Metadata objects
    """

    def setup(self):
        self.metadata = lc.data.metadata_pb2.Metadata()

    def time_initialize_metadata(self):
        lc.data.metadata_pb2.Metadata()

    def mem_metadata(self):
        return self.metadata


class BenchmarkLocData:
    """
    Benchmarks for initializing LocData objects
    """

    def setup(self):
        n_points = 1_000_000
        coordinates = rng.uniform(low=0, high=10_000, size=(4, n_points))
        locdata_dict = {
            "position_x": coordinates[0],
            "position_y": coordinates[1],
            "position_z": coordinates[2],
            "frame": np.linspace(0, n_points, n_points, dtype=int),
            "intensity": coordinates[3],
        }
        self.locdata_df = pd.DataFrame(locdata_dict)
        self.meta = lc.data.metadata_pb2.Metadata()
        self.meta.creation_time.seconds = 1
        self.locdata_3d = lc.LocData.from_dataframe(
            dataframe=self.locdata_df, meta=self.meta
        )

        n_selected_points = 1_000
        self.selection_indices = rng.integers(
            low=0, high=n_points, size=n_selected_points
        ).tolist()
        self.selection = lc.LocData.from_selection(
            locdata=self.locdata_3d, indices=self.selection_indices
        )

        self.locdatas = [
            lc.LocData.from_selection(locdata=self.locdata_3d, indices=idxs.tolist())
            for idxs in rng.integers(low=0, high=n_points, size=(10, n_selected_points))
        ]
        self.collection = lc.LocData.from_collection(locdatas=self.locdatas)

    def time_initialize_locdata_from_dataframe(self):
        lc.LocData.from_dataframe(dataframe=self.locdata_df, meta=self.meta)

    def time_locdata_selection(self):
        lc.LocData.from_selection(
            locdata=self.locdata_3d, indices=self.selection_indices
        )

    def time_locdata_collection(self):
        lc.LocData.from_collection(locdatas=self.locdatas)

    def mem_empty_locdata(self):
        return lc.LocData()

    def mem_locdata_from_dataframe(self):
        return self.locdata_3d

    def mem_locdata_selection(self):
        return self.selection

    def mem_locdata_selection_reduce(self):
        return self.selection.reduce()

    def mem_locdata_collection(self):
        return self.collection

    def mem_locdata_collection_reduce(self):
        return self.collection.reduce()


if __name__ == "__main__":
    class_ = BenchmarkLocData()
    class_.setup()
    print(class_.collection.data)



from pathlib import Path

import matplotlib.pyplot as plt  # this import is needed for visual inspection
import numpy as np
import pandas as pd
import pytest

from locan import ROOT_DIR, LocData
from locan.analysis import CoordinateBasedColocalization
from locan.analysis.cbc import _coordinate_based_colocalization
from locan.data.transform import transform_affine
from locan.locan_io.locdata.io_locdata import load_txt_file

# fixtures


@pytest.fixture()
def locdata():
    path = Path(ROOT_DIR / "tests/test_data/five_blobs.txt")
    dat = load_txt_file(path)
    return dat


@pytest.fixture()
def locdata_3d():
    path = Path(ROOT_DIR / "tests/test_data/five_blobs_3D.txt")
    dat = load_txt_file(path)
    return dat


@pytest.fixture()
def locdata_line():
    locdata_dict = {
        "position_x": [0, 1, 3, 4, 50, 98, 99, 100],
        "position_y": np.zeros(8),
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


def test_cbc(locdata):
    points = locdata.coordinates
    points_trans = transform_affine(locdata.coordinates)
    res = _coordinate_based_colocalization(
        points=points, other_points=points_trans, radius=100, n_steps=10
    )
    assert res[0] == 0.6162215398589203
    # print(res[0:5])

    # # plot data
    # fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10, 5))
    # axes[0].scatter(x=points[:,0], y=points[:,1], marker='.', c=res, cmap='jet', label='points')
    # axes[0].set_title('CBC')
    #
    # axes[1].scatter(x=points[:,0], y=points[:,1], marker='.', color='Blue', label='points')
    # axes[1].scatter(x=points_trans[:,0], y=points_trans[:,1], marker='o', color='Red', label='transformed points')
    # axes[1].set_title('Transformed points')
    #
    # plt.show()

    plt.close("all")


def test_cbc_nan(locdata_line):
    res = _coordinate_based_colocalization(
        points=locdata_line.coordinates,
        other_points=locdata_line.coordinates + (0.5, 0),
        radius=10,
        n_steps=10,
    )
    assert len(res) == 8
    assert np.isnan(res[4])


def test_Coordinate_based_colocalization_empty(caplog):
    cbc = CoordinateBasedColocalization().compute(LocData())
    cbc.hist()
    assert caplog.record_tuples == [("locan.analysis.cbc", 30, "Locdata is empty.")]

    plt.close("all")


def test_Coordinate_based_colocalization(locdata_line):
    other_locdata = transform_affine(locdata_line)
    cbc = CoordinateBasedColocalization(radius=100, n_steps=10).compute(
        locdata=locdata_line, other_locdata=other_locdata
    )
    # print(cbc.results)
    assert cbc.results.columns == f"colocalization_cbc_{other_locdata.meta.identifier}"

    cbc.hist()

    plt.close("all")



from locan import ROOT_DIR, find_pattern_upstream


def test_find_pattern_upstream():
    # find file
    sub_directory = ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt"
    pattern = "*.txt"
    pattern_found = find_pattern_upstream(sub_directory=sub_directory, pattern=pattern)
    assert pattern_found.parent == sub_directory.parent

    sub_directory = ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt"
    pattern = "__init__.py"
    pattern_found = find_pattern_upstream(sub_directory=sub_directory, pattern=pattern)
    assert pattern_found == ROOT_DIR / "tests/__init__.py"

    sub_directory = ROOT_DIR / "tests/test_data/"
    pattern = "__init__.py"
    pattern_found = find_pattern_upstream(sub_directory=sub_directory, pattern=pattern)
    assert pattern_found == ROOT_DIR / "tests/__init__.py"

    # find directory
    sub_directory = ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt"
    pattern = "tests"
    pattern_found = find_pattern_upstream(sub_directory=sub_directory, pattern=pattern)
    assert pattern_found == ROOT_DIR / "tests"

    # find no file
    sub_directory = ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt"
    pattern = "not_present.any"
    pattern_found = find_pattern_upstream(sub_directory=sub_directory, pattern=pattern)
    assert pattern_found is None

    sub_directory = ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt"
    pattern = "__init__.py"
    directory = ROOT_DIR / "tests/test_data"
    pattern_found = find_pattern_upstream(
        sub_directory=sub_directory, pattern=pattern, directory=directory
    )
    assert pattern_found is None



import numpy as np
import pytest

from locan import render_3d_napari, render_3d_rgb_napari, transform_affine
from locan.dependencies import HAS_DEPENDENCY

if HAS_DEPENDENCY["napari"]:
    import napari


@pytest.mark.visual
# this is to check overlay of rendered image and single localization points
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_napari_coordinates(locdata_blobs_3d):
    viewer, histogram = render_3d_napari(
        locdata_blobs_3d, bin_size=10, cmap="viridis", gamma=0.1
    )
    viewer.add_points(
        (locdata_blobs_3d.coordinates - np.array(histogram.bins.bin_range)[:, 0]) / 10
    )
    napari.run()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_napari(locdata_blobs_3d):
    render_3d_napari(locdata_blobs_3d, bin_size=100, cmap="viridis", gamma=0.1)
    # napari.run()

    viewer = render_3d_napari(locdata_blobs_3d, bin_size=50, cmap="magenta", gamma=0.1)
    render_3d_napari(
        locdata_blobs_3d,
        viewer=viewer,
        bin_size=100,
        cmap="cyan",
        gamma=0.1,
        scale=(2, 2),
        blending="additive",
    )
    napari.run()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_napari_empty(locdata_empty):
    render_3d_napari(locdata_empty, bin_size=100, cmap="viridis", gamma=0.1)
    napari.run()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_napari_single(locdata_single_localization, caplog):
    render_3d_napari(
        locdata_single_localization, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert caplog.record_tuples[1] == (
        "locan.render.render3d",
        30,
        "Locdata carries a single localization.",
    )
    napari.run()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_rgb_napari(locdata_blobs_3d):
    locdata_0 = locdata_blobs_3d
    locdata_1 = transform_affine(locdata_blobs_3d, offset=(20, 0, 0))
    render_3d_rgb_napari([locdata_0, locdata_1], bin_size=20)

    napari.run()



"""

This module contains tests that need gui interactions.

"""
import pytest

from locan import ROOT_DIR
from locan.dependencies import HAS_DEPENDENCY
from locan.gui.io import file_dialog


pytestmark = pytest.mark.qt


def test_file_dialog_no_qt():
    if not HAS_DEPENDENCY["qt"]:
        with pytest.raises(ImportError):
            file_dialog()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["qt"], reason="requires qt_binding.")
def test_file_dialog():
    file_path = ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt"
    result = file_dialog(directory=file_path, message="Select single file")
    # result = file_dialog()
    print(result)
    assert len(result) == 1
    assert result == str(file_path)


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["qt"], reason="requires qt_binding.")
def test_file_dialog():
    file_path = ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt"
    result = file_dialog(directory=file_path, message="Select single file")
    # result = file_dialog()
    print(result)
    assert len(result) == 1
    assert result == str(file_path)

##########################
@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["qt"], reason="requires qt_binding.")
def test_qtbot():
    from qtpy.QtWidgets import QApplication
    #app = QApplication([])
    directory = ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt"
    result = file_dialog(directory=directory, message="Select single file")
    #print(app.allWidgets())


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["qt"], reason="requires qt_binding.")
def test_qtbot_(qtbot):
    from qtpy.QtWidgets import QWidget, QApplication
    from qtpy.QtCore import Qt
    app = QApplication([])
    app.exec_()
    widget = QWidget()
    qtbot.addWidget(widget)
    qtbot.mouseClick(widget.findButton, Qt.LeftButton)


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["qt"], reason="requires qt_binding.")
def test_qtbot__(qtbot):
    from qtpy.QtWidgets import QWidget, QApplication
    from qtpy.QtCore import Qt
    app = QApplication([])
    directory = ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt"
    result = file_dialog(directory=directory, message="Select single file")
    widget = QApplication.activePopupWidget()
    print(widget)
    # qtbot.addWidget(widget)
    # qtbot.mouseClick(widget.findButton, Qt.LeftButton)



import logging
from io import BytesIO, StringIO, TextIOWrapper
from platform import system

import numpy as np
import pytest

from locan import ROOT_DIR
from locan.constants import DECODE_KEYS, THUNDERSTORM_KEYS
from locan.locan_io import convert_property_names, convert_property_types
from locan.locan_io.locdata.utilities import open_path_or_file_like


def test_open_path_or_file_like():
    if system() == "Linux":
        inputs = [
            str(ROOT_DIR) + "/tests/test_data/rapidSTORM_dstorm_data.txt",
            ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt",
        ]
    elif system() == "Windows":
        inputs = [
            str(ROOT_DIR) + "/tests/test_data/rapidSTORM_dstorm_data.txt",
            str(ROOT_DIR) + r"\tests\test_data\rapidSTORM_dstorm_data.txt",
            ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt",
            ROOT_DIR / r"tests\test_data\rapidSTORM_dstorm_data.txt",
        ]
    else:
        inputs = [
            str(ROOT_DIR) + "/tests/test_data/rapidSTORM_dstorm_data.txt",
            ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt",
        ]

    for pfl in inputs:
        with open_path_or_file_like(path_or_file_like=pfl) as data:
            assert isinstance(data, TextIOWrapper)
            out = data.read(10)
            assert out and isinstance(out, str)
            assert not data.closed
        assert data.closed

    pfl = StringIO("This is a file-like object to be read.")
    with open_path_or_file_like(path_or_file_like=pfl) as data:
        assert isinstance(data, StringIO)
        out = data.read(10)
        assert out and isinstance(out, str)
        assert not data.closed
    assert data.closed

    pfl = BytesIO(b"This is a file-like object to be read.")
    with open_path_or_file_like(path_or_file_like=pfl) as data:
        assert isinstance(data, BytesIO)
        out = data.read(10)
        assert out and isinstance(out, bytes)
        assert not data.closed
    assert data.closed

    pfl = ROOT_DIR / "tests/test_data/some_file_that_does_not_exits.txt"
    with pytest.raises(FileNotFoundError):
        with open_path_or_file_like(path_or_file_like=pfl):
            pass


def test_convert_property_types(locdata_2d):
    df = locdata_2d.data.copy()
    types_mapping = {
        "position_x": "float",
        "position_y": str,
        "frame": np.int64,
        "not_in_there": "float",
    }
    converted_df = convert_property_types(
        dataframe=df, loc_properties=None, types=types_mapping
    )
    assert converted_df.dtypes[0] == np.float32
    assert converted_df.dtypes[1] == object
    assert converted_df.dtypes[2] == np.int64
    assert converted_df.dtypes[3] == np.int32
    assert isinstance(converted_df["position_y"].iloc[0], str)


def test_convert_property_names(caplog):
    new_properties = convert_property_names(
        properties=["position_x"], property_mapping=None
    )
    assert new_properties == ["position_x"]

    properties = ["position_x", "y", "z", "frame_ix", "something"]
    new_properties = convert_property_names(
        properties=properties, property_mapping=None
    )
    assert new_properties == ["position_x", "y", "z", "frame_ix", "something"]
    assert caplog.record_tuples[0] == (
        "locan.locan_io.locdata.utilities",
        logging.WARNING,
        "Column y is not a Locan property standard.",
    )

    new_properties = convert_property_names(
        properties=properties, property_mapping=DECODE_KEYS
    )
    assert new_properties == [
        "position_x",
        "position_y",
        "position_z",
        "frame",
        "something",
    ]
    assert caplog.record_tuples[4] == (
        "locan.locan_io.locdata.utilities",
        logging.WARNING,
        "Column something is not a Locan property standard.",
    )

    properties = ["position_x", "y", "z", "frame_ix", "something", "sigma [nm]"]
    new_properties = convert_property_names(
        properties=properties, property_mapping=[DECODE_KEYS, THUNDERSTORM_KEYS]
    )
    assert new_properties == [
        "position_x",
        "position_y",
        "position_z",
        "frame",
        "something",
        "psf_sigma",
    ]
    assert caplog.record_tuples[4] == (
        "locan.locan_io.locdata.utilities",
        logging.WARNING,
        "Column something is not a Locan property standard.",
    )



"""

This module contains tests that need gui interactions.

"""
import pytest

from locan import ROOT_DIR
from locan.dependencies import HAS_DEPENDENCY
from locan.gui.io import file_dialog

pytestmark = pytest.mark.qt


def test_file_dialog_no_qt():
    if not HAS_DEPENDENCY["qt"]:
        with pytest.raises(ImportError):
            file_dialog()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["qt"], reason="requires qt_binding.")
def test_file_dialog():
    file_path = ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt"
    result = file_dialog(directory=file_path, message="Select single file")
    # result = file_dialog()
    # print(result)
    assert len(result) == 1
    assert result == str(file_path).replace("\\", "/") or result[0] == str(
        file_path
    ).replace("\\", "/")
